@@ -62,8 +62,7 @@ static int quiet = 0;
 static int batch_mode = 0;
 static char *query_file = NULL;
 static int json = 0;
-static char *the_root = ".";
-static char *name;
+static char name[2048] = ".";
 static getdns_context *context;
 static getdns_dict *extensions;
 static getdns_dict *query_extensions_spc = NULL;
@@ -74,9 +73,11 @@ static getdns_dict *listen_dict = NULL;
 static size_t pincount = 0;
 static size_t listen_count = 0;
 static uint16_t request_type = GETDNS_RRTYPE_NS;
+static int got_rrtype = 0;
 static int timeout, edns0_size, padding_blocksize;
 static int async = 0, interactive = 0;
 static enum { GENERAL, ADDRESS, HOSTNAME, SERVICE } calltype = GENERAL;
+static int got_calltype = 0;
 static int bogus_answers = 0;
 static int check_dnssec = 0;
 #ifndef USE_WINSOCK
@@ -97,7 +98,7 @@ static int get_rrtype(const char *t)
 	if (strlen(t) > sizeof(buf) - 15)
 		return -1;
 	for (i = 14; *t && i < sizeof(buf) - 1; i++, t++)
-		buf[i] = *t == '-' ? '_' : toupper(*t);
+		buf[i] = *t == '-' ? '_' : toupper((unsigned char)*t);
 	buf[i] = '\0';
 
 	if (!getdns_str2int(buf, &rrtype))
@@ -122,7 +123,7 @@ static int get_rrclass(const char *t)
 	if (strlen(t) > sizeof(buf) - 16)
 		return -1;
 	for (i = 15; *t && i < sizeof(buf) - 1; i++, t++)
-		buf[i] = toupper(*t);
+		buf[i] = toupper((unsigned char)*t);
 	buf[i] = '\0';
 
 	if (!getdns_str2int(buf, &rrclass))
@@ -581,8 +582,6 @@ getdns_return_t parse_args(int argc, char **argv)
 	size_t upstream_count = 0;
 	FILE *fh;
 	int int_value;
-	int got_rrtype = 0;
-	int got_calltype = 0;
 	int got_qname = 0;
 
 	for (i = 1; i < argc; i++) {
@@ -658,8 +657,15 @@ getdns_return_t parse_args(int argc, char **argv)
 			continue;
 
 		} else if (arg[0] != '-') {
+			size_t arg_len = strlen(arg);
+
 			got_qname = 1;
-			name = arg;
+			if (arg_len > sizeof(name) - 1) {
+				fprintf(stderr, "Query name too long\n");
+				return GETDNS_RETURN_BAD_DOMAIN_NAME;
+			}
+			(void) memcpy(name, arg, arg_len);
+			name[arg_len] = 0;
 			continue;
 		}
 		for (c = arg+1; *c; c++) {
@@ -1181,7 +1187,7 @@ getdns_return_t do_the_call(void)
 			r = GETDNS_RETURN_GENERIC_ERROR;
 			break;
 		}
-		if (r == GETDNS_RETURN_GOOD && !batch_mode) 
+		if (r == GETDNS_RETURN_GOOD && !batch_mode && !interactive) 
 			getdns_context_run(context);
 		if (r != GETDNS_RETURN_GOOD)
 			fprintf(stderr, "An error occurred: %d '%s'\n", (int)r,
@@ -1258,14 +1264,28 @@ static void incoming_request_handler(getdns_context *context,
     void *userarg, getdns_transaction_t request_id);
 
 
+void read_line_cb(void *userarg);
+void read_line_tiny_delay_cb(void *userarg)
+{
+	getdns_eventloop_event *read_line_ev = userarg;
+
+	loop->vmt->clear(loop, read_line_ev);
+	read_line_ev->timeout_cb = NULL;
+	read_line_ev->read_cb = read_line_cb;
+	loop->vmt->schedule(loop, fileno(fp), -1, read_line_ev);
+}
+
 void read_line_cb(void *userarg)
 {
+	static int n = 0;
 	getdns_eventloop_event *read_line_ev = userarg;
 	getdns_return_t r;
 
 	char line[1024], *token, *linev[256];
 	int linec;
 
+	assert(n == 0);
+	n += 1;
 	if (!fgets(line, 1024, fp) || !*line) {
 		if (query_file && verbosity)
 			fprintf(stdout,"End of file.");
@@ -1276,6 +1296,7 @@ void read_line_cb(void *userarg)
 		if (interactive && !query_file)
 			(void) getdns_context_set_upstream_recursive_servers(
 			    context, NULL);
+		n -= 1;
 		return;
 	}
 	if (query_file && verbosity)
@@ -1288,6 +1309,7 @@ void read_line_cb(void *userarg)
 			printf("> ");
 			fflush(stdout);
 		}
+		n -= 1;
 		return;
 	}
 	if (*token == '#') {
@@ -1297,6 +1319,7 @@ void read_line_cb(void *userarg)
 			printf("> ");
 			fflush(stdout);
 		}
+		n -= 1;
 		return;
 	}
 	do linev[linec++] = token;
@@ -1312,10 +1335,23 @@ void read_line_cb(void *userarg)
 	    (r != CONTINUE && r != CONTINUE_ERROR))
 		loop->vmt->clear(loop, read_line_ev);
 
-	else if (! query_file) {
-		printf("> ");
-		fflush(stdout);
+	else {
+#if 0
+		/* Tiny delay, to make sending queries less bursty with
+		 * -F parameter.
+		 *
+		 */
+		loop->vmt->clear(loop, read_line_ev);
+		read_line_ev->read_cb = NULL;
+		read_line_ev->timeout_cb = read_line_tiny_delay_cb;
+		loop->vmt->schedule(loop, fileno(fp), 1, read_line_ev);
+#endif
+		if (! query_file) {
+			printf("> ");
+			fflush(stdout);
+		}
 	}
+	n -= 1;
 }
 
 typedef struct dns_msg {
@@ -1740,7 +1776,6 @@ main(int argc, char **argv)
 {
 	getdns_return_t r;
 
-	name = the_root;
 	if ((r = getdns_context_create(&context, 1))) {
 		fprintf(stderr, "Create context failed: %d\n", (int)r);
 		return r;
